/** Plogg, a curse(s/d) mp3/Ogg player, based on original code
 * from "jb - a curses based jukebox for mp-compressed audio"
 * (C) 1997 Kristian Wiklund (under the GPL).
 * $Id: plogg.c 1.1 2002/09/16 02:27:50 KILGORE Exp KILGORE $
 *
 * (C) 2001 Jeff Robinson (jeffnik@anecho.mb.ca), placed under the GPL
 * The Gnu General Public License as described below is available
 * in the file COPYING distributed with this package.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include <os2.h>
#include <stdio.h>
#define VERSION "2003-03-16"

#define INCL_DOS

#include <curses.h>

#include <process.h>
#include <signal.h>
#include <assert.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#ifdef HAVE_SYS_FILE_H
#include <sys/file.h>
#endif
// #ifdef STDC_HEADERS
#include <string.h>
#include <stdlib.h>
// #else
// # ifndef HAVE_MEMCPY
// #  define memcpy(d, s, n) bcopy ((s), (d), (n))
// #  define memmove(d, s, n) bcopy ((s), (d), (n))
// # endif
// #endif

#ifdef HAVE_SYS_SOUNDCARD_H
#define LINUX_SOUND
#include <sys/soundcard.h>
#endif

#include "plogg.h"
// #include "version.h"

/* globals:
 */

char *musiclist="list"; /* Name of our playlist */
char oggPath[256]="ogg123.exe";
char oggArgs[256]="-q -ddart";
char mpgPath[256]="mpg123.exe";
char mpgArgs[256]="-q";

char *mainpath="";

struct tune {
  char *path;
  int time, status;
} *tunes;

WINDOW *top, *middle, *bottom;
int currentType = 0;          /* Current file-type being played */
int oneShotPlaylist = 0;        /* A non-zero value means we have been passed files to play */
static const int NONE = 0;
static const int OGG = 1;
static const int MP3 = 2;

int *playlist = 0;
int pllen = 0, plmlen = 0;
int nrsongs = 0, nrtagged = 0;
int tunenr = 0; /* the currently selected tune */
int msize = 0;  /* size of displayed tunelist */
int startedtune = -1, toptune = 0; /* the tune at the top of middle window */
#ifdef LINUX_SOUND
int volume, mixer_fd;
#endif

#define TAGGED 1

pid_t playerpid = 0;
int starttime;


static void stop_playing() {

    signal(playerpid, SIG_ACK);
    kill(playerpid, 9); /* yeah! BFG! */

    // Under heavy system load, sometimes it seems like we don't
    // actually launch a new player.  Should we Fix this XXX?

    playerpid = 0;
    currentType = NONE;
}

/* if the initial string of char s matches mainpath we
 * remove mainpath from s by adding strlen(mainpath) to
 * s
 */
static char *modify_name(char *s) {
    int l = strlen(mainpath);
    static char buf[1024];
    static char tmp[80];

    if(!s) {
        return 0;
    }

    if(!strncmp(mainpath, s, l)) {
        strcpy(buf, s+l);
    } else {
        strcpy(buf, s);
    }

    if(strlen(buf)>(COLS-11)) {
        buf[strlen(buf)-11]=0;
    }

    if (strlen(buf) > (COLS - 11)) {
        /* Hey look!  Ugly code!  Fix this XXX ! */
        strncpy(tmp, buf, (COLS - 11));
        strcpy(buf, tmp);
    }

    return buf;
}

RETSIGTYPE dostuff(int i) {
    int status;
    int ec;

    waitpid(playerpid, &status, 0);

    ec = WEXITSTATUS(status);

    /* this could be used to determine the play time of
     * the tune. also, check WIFEXITED, to see if user terminated
     * the player
     */
    if(WIFEXITED(status) && ec==0) {
        tunes[startedtune].time = time(0) - starttime;
    }

    playerpid = 0;
    wclear(top);
    waddstr(top, "<nothing>");
    mvwaddstr(top, 0, COLS-6, " 0:00");
    wmove(top, 1, 0);
    for(i=0;i<COLS;i++) {
        waddch(top, '-');
    }
    wrefresh(top);

    // signal(SIGCHLD, dostuff);
    // signal(SIGCHLD, SIG_ACK);
    signal((int)dostuff, SIG_ACK);
    // signal(i, SIG_ACK);
    // signal(SIGCHLD, SIG_ACK);
}

static void finish(int sig)
{
  FILE *f;
  int i, fd;
#if !defined(HAVE_FLOCK) && !defined(HAVE_LOCKF)
  flock_t flocks;
  #endif
  char buf[1024];

  endwin();

  /* if we don't do this and kill the player while
   * working on a playlist, we get the next tune played
   * when quitting
   */
  // signal(SIGCHLD, SIG_IGN);
  signal(playerpid, SIG_IGN);

  if(playerpid) {
      kill(playerpid, SIGTERM);
  }

  /* don't bother doing the save things if user can't access
   * the list!
   */
  // if(access(MUSICLIST, W_OK)) {
  if (access(musiclist, W_OK)) {
      exit(0);
  }

  if (oneShotPlaylist > 0) {
      exit(0);
  }

  sprintf(buf, "%s.old", musiclist);
  rename(musiclist, buf);

  printf("Saving music list...");
  fd = open(musiclist,O_RDONLY);

#ifdef HAVE_LOCKF
  lockf(fd, F_LOCK, 0);
#else
#ifdef HAVE_FLOCK
  flock(fd, LOCK_EX);
#else
  flocks.l_type = F_WRLCK;
  flocks.l_len = 0;
  fcntl(fd, F_SETLKW, &flocks);
#endif
#endif

  assert(f = fopen(musiclist, "w"));

  // for(i=0;i<nrsongs-1;i++) {
  for(i=0; i < nrsongs; i++) {
      fprintf(f, "%s#%d\n", tunes[i].path, tunes[i].time);
      if(!(i%10)) {
          printf(".");
          // fsync(0);  // This call may be potentially dangerous on older OS/2 systems (even Warp 3) using HPFS?!
      }
  }

  fclose(f);

#ifdef HAVE_LOCKF
  lockf(fd, F_ULOCK, 0);
#else
#ifdef HAVE_FLOCK
  flock(fd, LOCK_UN);
#else
  flocks.l_type = F_UNLCK;
  fcntl(fd, F_SETLK, &flocks);
#endif
#endif

  puts("");
  close(fd);
  exit(0);
}

/** Display the elapsed time of the currently playing song */
static void tunetime(int i) {
    char buf[20];

    // Calculate the playing time, if playing
    if(tunes[toptune+i].time) {
        sprintf(buf, "%2d:%02d", tunes[toptune+i].time/60,
                tunes[toptune+i].time%60);
        mvwaddstr(middle, i, 1, buf);
    } else {
        mvwaddstr(middle, i, 1, "     ");
    }

    // Show a '+' symbol next to the playing song on the playlist
    if(tunes[toptune+i].status&TAGGED) {
        mvwaddch(middle, i, 0, '+');
    } else {
        mvwaddch(middle, i, 0, ' ');
    }
}

static void refresh_screen() {
    int i;
    //    char title[256];

    wclear(top);
    wclear(middle);
    wclear(bottom);

    for(i=0; i<msize; i++)
    {
        if(i+toptune >= nrsongs) {
            break;
        }

        if(i+toptune == tunenr) {
            wattron(middle, COLOR_PAIR(3));
            mvwaddstr(middle, i, 7, modify_name(tunes[i + toptune].path));
            wattroff(middle, COLOR_PAIR(3));
            tunetime(i);
        } else {
            mvwaddstr(middle, i, 7, modify_name(tunes[i + toptune].path));
            tunetime(i);
        }
    }

#ifndef NoDLLs
    // This next line is only for testing purposes.  Fix me XXX
    openOgg("f:/keoki_passiton.ogg");
    // openOgg("e:\\bin\\audio\\mp3\'s\\keoki_passiton.ogg");

    // Get the song's title via muzthunk.dll
    getSongTitle(&title);

    // We'd better put some bounds checking and stuff in here, too!  Fix me XXX
    waddstr(top, title);
#else
    wattron(top, A_BOLD);

    if(!playerpid) {
        wattron(top,COLOR_PAIR(1));
        waddstr(top, "<nothing is playing>");
        wattroff(top, COLOR_PAIR(1));
    } else {
        wattron(top, COLOR_PAIR(2));
        waddstr(top, modify_name(tunes[startedtune].path));
        wattroff(top, COLOR_PAIR(2));
    }

    wattroff(top, A_BOLD);
#endif

    wattron(top, COLOR_PAIR(1));
    mvwaddstr(top, 0, COLS-6, " 0:00");
    wattroff(top, COLOR_PAIR(1));
    wmove(top, 1, 0);
    for(i=0;i<COLS;i++)
        waddch(top, '-');

#ifdef LINUX_SOUND
    mvwaddch(top, 1, volume*(COLS-1)/128, '*');
#endif

    for(i=0;i<COLS;i++) {
        waddch(bottom, '-');
    }

    wattron(bottom, A_BOLD);
    wattron(bottom, COLOR_PAIR(1));
    mvwaddstr(bottom, 1, 0,
              "arrow keys to navigate, t - toggle selection, p - play tune, v - view playlist");
    // mvwaddstr(bottom, 2, 0, "g - play selected, ~ - select/deselect all, q - quit, h - help");
    mvwaddstr(bottom, 2, 0, "g - play selected, ~ - select/deselect all, q - quit playing, ESC - exit");
    wattroff(bottom, COLOR_PAIR(1));
    wattroff(bottom, A_BOLD);

    wrefresh(top);
    wrefresh(middle);
    wrefresh(bottom);
}

static void draw_list() {
    int i;

    // Check to see if we need to scroll down the list
    // while (tunenr - toptune >= msize) {
    while (tunenr - toptune >= msize) {
        toptune++;
        i = toptune + msize - 1;
        scroll(middle);

        if ((tunes[i].status & TAGGED) > 0) {
            mvwaddch(middle, msize - 1, 0, '+');
        }
        mvwaddstr(middle, msize - 1, 7, modify_name(tunes[i].path));
        wrefresh(middle);
    }


    // Check if we need to scroll up the list
    while (tunenr - toptune < 0) {
        toptune--;
        wscrl(middle, -1);

        if ((tunes[toptune].status & TAGGED) > 0) {
            mvwaddch(middle, 0, 0, '+');
        }

        mvwaddstr(middle, 0, 7, modify_name(tunes[toptune].path));
        wrefresh(middle);
    }

    i = tunenr - toptune;

    // wstandout(middle);

    // We have to pay a little more attention to the centre portion
    // as the title may be too long and cause ugly wrap.  Right now
    // we'll hard-code this for a 80-character display width, but
    // in the future we should do this in a more elegant way.
    wattron(middle, COLOR_PAIR(3));
    mvwaddstr(middle, i, 7, modify_name(tunes[tunenr].path));
    wattroff(middle, COLOR_PAIR(3));

    tunetime(i);
    // wstandend(middle);

    wrefresh(middle);
  
}

RETSIGTYPE alarmhandler(int sig) {
    int status;
    int ec;
    int diff;
    char buf[20];

    if(!playerpid) {
        return;
    }

    /* Test to see if the player is still running */
    ec = waitpid(playerpid, &status, WNOHANG);

    if (ec == -1) {
        // The player isn't there anymore, so let's make certain
        // that we stop the timer and move on.

        // Should the be playnextselected() instead?  Sometimes it stops dead.  Fix me XXX !
        // stop_playing();
        playnextselected();
    }

    diff = time(0) - starttime;

    /* if we know how long the tune is, show remaining time as well.
     */

    // attron(COLOR_PAIR(1));
    wattron(top, A_BOLD);
    wattron(top, COLOR_PAIR(2));
    if(tunes[startedtune].time>0) {
        sprintf(buf, "%2d:%02d", (tunes[startedtune].time-diff)/60, (tunes[startedtune].time-diff)%60);
        mvwaddstr(top, 0, COLS-12, buf);
    }
    sprintf(buf, "%2d:%02d", diff/60, diff%60);
    // attroff(COLOR_PAIR(1));
    // testing with colour
    // init_pair(1, COLOR_RED, COLOR_YELLOW);
    // wattrset(top, COLOR_PAIR(1));
    mvwaddstr(top, 0, COLS-6, buf);
    wattroff(top, COLOR_PAIR(2));
    wattroff(top, A_BOLD);
    wrefresh(top);
    signal(SIGALRM, alarmhandler);
    // signal(sig, SIG_ACK);
    // signal(SIGALRM, SIG_ACK);
    // signal(alarmhandler, SIG_ACK);
    alarm(1);
}

int selectptr = 0;

static void playnextselected();

RETSIGTYPE selplay(int sig) {
    int status;
    int ec;

    waitpid(playerpid, &status, 0);

    ec = WEXITSTATUS(status);

    playerpid = 0;

    playnextselected();


}

static void playnextselected() {
    int i;
    int intunenr;
    int ec, status;

    if(selectptr==pllen) {
        stop_playing();

        // If this is just a one-shot (list of songs from command line/drag & drop)
        if (oneShotPlaylist > 0) {
            exit(0);
        }

        // signal(SIGALRM, alarmhandler);
        refresh_screen();
        return;
    }

    intunenr  = playlist[selectptr];
    selectptr++;

    if (!playerpid) {

        // If we don't wait and make certain that the previous player stopped,
        // then the program just cycles on through the rest of the playlist.
        // There may be a better way than just using "wait()", though...
        // this could be a potential "hang"

        waitpid(playerpid, &status, 0);

        ec = WEXITSTATUS(status);
        if (ec == 0) {
            wait(playerpid);
        }
    }

    choosePlayer(intunenr);

    // startedtune = tunenr;
    startedtune = intunenr;

    starttime = time(0);
    wclear(top);
    // waddstr(top, modify_name(tunes[tunenr].path));
    wattron(top, A_BOLD);
    wattron(top, COLOR_PAIR(2));
    waddstr(top, modify_name(tunes[intunenr].path));
    wattroff(top, COLOR_PAIR(2));
    wattroff(top, A_BOLD);
    mvwaddstr(top, 0, COLS-6, " 0:00");
    wmove(top, 1, 0);
    for(i=0;i<COLS;i++) {
        waddch(top, '-');
    }

    wrefresh(top);
    signal(SIGALRM, alarmhandler);
    alarm(1);
    // signal(SIGCHLD, selplay);
    signal((int)selplay, SIG_ACK);
}

static void make_playlist()
{
  int i;
  
  selectptr = 0;

  if(playlist && plmlen<nrtagged)
    {
      free(playlist);
      playlist = 0;
    }

  if(!playlist)
    {
      playlist = (int *)malloc(sizeof(int)*nrtagged*2);
      plmlen = nrtagged*2;
  }
  
  for(i=0;i<nrsongs;i++)
    {
      if(tunes[i].status&TAGGED)
	playlist[selectptr++] = i;
    }
  selectptr = 0;  
  pllen = nrtagged;

}

static void add_to_playlist(int i)
{
  if(plmlen<=pllen+1)
    {
      if(plmlen==0)
	plmlen=256;

      assert(playlist = (int *)realloc(playlist,sizeof(int)*plmlen*2));
      plmlen*=2;
    }
  playlist[pllen++] = i;
}

static void make_random_playlist() {
    /* Fix this XXX */

    int i, x, y, t;
    make_playlist();

    for(i=0;i<4*pllen;i++) {
        // x = random() % pllen;
        // y = random() % pllen;
        x = rand() % pllen;
        y = rand() % pllen;
        t = playlist[x];
        playlist[x] = playlist[y];
        playlist[y] = t;
    }

}

#ifdef LINUX_SOUND

static void volume_down()
{
  int i;

  mvwaddch(top, 1, volume*(COLS-1)/128, '-');
  volume--;
  if(volume<0)
    volume=0;
  mvwaddch(top, 1, volume*(COLS-1)/128, '*');
  wrefresh(top);
  i = volume | volume<<8;
  ioctl(mixer_fd, MIXER_WRITE(SOUND_MIXER_VOLUME), &i);
}

static void volume_up()
{
  int i;

  mvwaddch(top, 1, volume*(COLS-1)/128, '-');
  volume++;
  if(volume>128)
    volume=128;
  mvwaddch(top, 1, volume*(COLS-1)/128, '*');
  wrefresh(top);
  i = volume | volume<<8;
  ioctl(mixer_fd, MIXER_WRITE(SOUND_MIXER_VOLUME), &i);
}

#endif

static void view_playlist() {
    int i, sl;
    int c;
    WINDOW *w,*l;
    /* pop up a huge window above the middle and bottom ones,
     * then draw the current playlist in it. provide functions for
     * reordering, loading and saving
     */

    w = newwin(LINES-6, 0, 2, 0);
    l = newwin(4, 0, LINES-4, 0);
    wclear(w);
    wclear(l);

    for (i=0; i<COLS; i++) {
        waddch(l, '-');
    }

    mvwaddstr(l,1,0,
              "ESC - return to main screen");

    // Now draw the playlist
    int startlist = selectptr - (msize / 2);
    if (startlist < 0) {
        startlist = 0;
    }

    if (startlist + msize > nrsongs) {
        startlist = startlist - nrsongs;
    }

    if(playlist) {
        for (i=0; i<msize; i++) {
            if(i>=pllen) {
                break;
            }

            sl = i + startlist;

            // if(!i) {
            if (sl + 1 == selectptr) {
                wattron(w, A_BOLD);
            }

            if(tunes[playlist[sl]].time) {
                mvwprintw(w, i, 1, "%2d:%02d %s",
                          tunes[playlist[sl]].time/60,
                          tunes[playlist[sl]].time%60,
                          modify_name(tunes[playlist[sl]].path));
            } else {
                wattron(w, COLOR_PAIR(2));
                mvwaddstr(w, i, 6, modify_name(tunes[playlist[sl]].path));
                wattroff(w, COLOR_PAIR(2));
            }

            // if(!i) {
            if (sl + 1 == selectptr) {
                wattroff(w, A_BOLD);
            }
        }
    }

    wrefresh(w);
    wrefresh(l);

    while(1) {
        c = getch();

        switch(c) {
        case 27:
            // case 'Q':
            // case 'q':
            // goto getout;
            goto leavelist;
        // case 'p':
        case 'g':             // This was: case 'P':
            {
                // signal(SIGCHLD, selplay);
                signal((int)selplay, SIG_ACK);

                /* start tune only if nothing is playing, we want the
                 * user to be able to finish listening to the current song
                 */

                selectptr = 0;

                if(!playerpid) {
                    playnextselected();
                }
            }
            break;
        case 'Q':
        case 'q':
            //case 27:
            stop_playing();
            break;
        }
    }

getout:
    // Make sure our players are toast
    // if (playerpid) {
    // If we don't send a SIG_IGN, the program tries to play one more song!!
    // pllen = 0; // Set the length of the playlist to zero
    signal(playerpid, SIG_IGN);
    kill(playerpid, 9);
    signal(playerpid, SIGTERM);
    delwin(w);
    delwin(l);
    refresh_screen();

    exit(0);


leavelist:
    // Clear the window and refresh the screen
    delwin(w);
    delwin(l);
    refresh_screen();
}

static void action(int c) {
    int i;

    switch(c) {
    case 12:
        refresh_screen();
        break;
        // case '?':
    // case 'h':
        // system(READMECMD);
        // refresh_screen();
        // break;

#ifdef LINUX_SOUND
    case 'h':
    case KEY_LEFT:
        volume_down();
        break;

    case 'l':
    case KEY_RIGHT:
        volume_up();
        break;
#endif
    case 'a':
        add_to_playlist(tunenr);
        break;

    case 'q':
        /* stop playing what we are playing if we press escape! */
        signal(SIGCHLD, SIG_IGN);
        signal(playerpid, SIG_IGN);
        stop_playing();
        break;
    case 'v': /* view the current playlist */
        {
            view_playlist();
        }
        break;

    case '~':
        {
            if(tunes[0].status & TAGGED)
            {
                for(i=0;i<nrsongs;i++)
                    tunes[i].status&=~TAGGED;
                nrtagged=0;
            }
            else
            {
                for(i=0;i<nrsongs;i++)
                    tunes[i].status|=TAGGED;
                nrtagged=nrsongs;
            }
            refresh_screen();
            break;
        }
    case 27:
        // case 'q':
        // case 'Q':
        signal(SIGCHLD, SIG_IGN);
        signal(playerpid, SIG_IGN);
        finish(0);
        // exit(0);
        break;
    case 'j':
    case KEY_DOWN:
        i = tunenr - toptune;
        mvwaddstr(middle, i, 7, modify_name(tunes[i + toptune].path));
        tunetime(i);
        tunenr++;
        // if(tunenr>=nrsongs-2)
        // tunenr=nrsongs-2;
        if (tunenr >= nrsongs - 1) {
            tunenr = nrsongs -1;
        }

        draw_list();
        wrefresh(middle);
        break;
    case 'k':
    case KEY_UP:
        {
            i = tunenr - toptune;
            mvwaddstr(middle, i, 7, modify_name(tunes[i + toptune].path));
            tunetime(i);
            tunenr--;
            if(tunenr<0) {
                tunenr=0;
            }
            draw_list();
            wrefresh(middle);
            break;
        }

    case KEY_NPAGE:
        {
            i = tunenr - toptune;
            mvwaddstr(middle, i, 7, modify_name(tunes[i + toptune].path));
            tunetime(i);
            tunenr = tunenr + msize;
            // tunenr++;
            if (tunenr>=nrsongs - 2) {
                tunenr=nrsongs - 2;
            }
            draw_list();
            wrefresh(middle);
            break;
        }

    case KEY_PPAGE:
        {
            // Move up the list one "screen-length"
            i = tunenr - toptune;
            mvwaddstr(middle, i, 7, modify_name(tunes[i + toptune].path));
            tunetime(i);
            // tunenr--;
            tunenr = tunenr - msize;
            if(tunenr<0) {
                tunenr=0;
            }
            draw_list();
            wrefresh(middle);
            break;
        }

    case KEY_END:
        {
            tunenr = nrsongs - 1;
            draw_list();
            wrefresh(middle);
            break;
        }

    case KEY_HOME:
        {
            tunenr = 0;
            draw_list();
            wrefresh(middle);
            break;
        }

        // case 'n':
    case KEY_LEFT:
        {
            // A nasty hack to play the previous song
            selectptr = selectptr - 2;
            // signal(SIGCHLD, SIG_IGN);
            // signal(playerpid, SIG_IGN);
            // stop_playing();

            // Brute force.  Ugly, but effective!
            kill(playerpid, 9);
            break;
        }

    case KEY_RIGHT:
        {
            // skip to the next song
            // stop_playing();
            // playnextselected();
            // signal(SIGCHLD, SIG_IGN);
            // signal(playerpid, SIG_IGN);
            // stop_playing();

            // This is sort've a brute-force method, but seems to do the trick!
            kill(playerpid, 9);
            // wait(playerpid);
            break;
        }

    case 'p':
        /* we must spawn an mpg123-player, if no player is running,
         * but also zap the running player, if one is running
         */

        playSelectedSong();

        /*
        if(playerpid) {
            signal(SIGCHLD, SIG_IGN);
            signal(playerpid, SIG_IGN);
            stop_playing();
        }

        choosePlayer(tunenr);

        startedtune = tunenr;

        starttime = time(0);
        wclear(top);
        wattron(top, COLOR_PAIR(2));
        wattron(top, A_BOLD);
        waddstr(top, modify_name(tunes[tunenr].path));
        wattroff(top, A_BOLD);
        mvwaddstr(top, 0, COLS-6, " 0:00");
        wattroff(top, COLOR_PAIR(2));
        wmove(top, 1, 0);
        for(i=0;i<COLS;i++)
            waddch(top, '-');

        wrefresh(top);
        signal(SIGALRM, alarmhandler);
        // signal(alarmhandler, SIG_ACK);
        alarm(1);
        */

        break;
    case 't': /* tag the current tune */
        {
            int i = (tunes[tunenr].status & TAGGED)>0;

            i = i^1;

            wstandout(middle);
            if(i)
            {
                tunes[tunenr].status|=TAGGED;
                mvwaddch(middle, tunenr-toptune, 0, '+');
                nrtagged++;
            }
            else
            {
                tunes[tunenr].status&=~TAGGED;
                mvwaddch(middle, tunenr-toptune, 0, ' ');
                nrtagged--;
            }
            wstandend(middle);
            wrefresh(middle);
        }
        break;
    case 'g': /* play selected music */
        {
            // int c;
            // int quit = 0;

            signal(SIGCHLD, selplay);
            // signal (selplay, SIG_ACK);

            /* start tune only if nothing is playing, we want the
             * user to be able to finish listening to the current song
             */

            selectptr = 0;
            make_playlist();


            if(!playerpid) {
                playnextselected();
            }

        }
        break;
    case 'r': /* randomize over tunes */

        srand( (unsigned)time( NULL ) );
        {
            // int quit = 0;

            signal(SIGCHLD, selplay);
            // signal(selplay, SIG_ACK);

            /* start tune only if nothing is playing, we want the
             * user to be able to finish listening to the current song
             */

            selectptr = 0;
            make_random_playlist();

            if(!playerpid) {
                playnextselected();
            }
        }
        break;
    }
}

static void make_ui() {
    // int i;

    /* create a three-window ui:
     * top line shows currently playing tune, and elapsed time.
     * if tune was played before, show remaining time as well.
     *
     * then comes a tune selection window.
     *
     * at the bottom, we want a menu window.
     * (looks sort of like the elm interface.)
     */
    top = newwin(2, 0, 0, 0);
    middle = newwin(LINES-6, 0, 2, 0);
    bottom = newwin(4, 0, LINES-4, 0);

    msize = LINES-6;

    scrollok(middle, TRUE);
}

int main(int argc, char **argv) {
    int c, i;
    char a;

    while((a=getopt(argc, argv, "hv"))>-1) {
        switch(a)
        {
	case 'h': fprintf(stderr, "USAGE: %s [-h] [-v]\n", argv[0]);
        exit(0);
        case 'v':
            {
                fprintf(stderr, "'Plogg' "VERSION" (C) 2002 Jeff Robinson (jeffnik@anecho.mb.ca>\n");
                fprintf(stderr, "Based on source code from Jukebox by K. Wiklund.\n");
                fprintf(stderr, "'Jukebox' (C) K. Wiklund 1997 <kw@dtek.chalmers.se>\n");
                exit(0);
            }
        case '?':
            exit(1);
        }
    }

    initscr();
    // Test to see if our terminal supports colours or not
    if(has_colors() == FALSE) {
        endwin();
        printf("You terminal does not support colour\n");
        exit(1);
    }

    // "Initialise" colours and define our colour pairs
    start_color();
    init_color(COLOR_YELLOW, 1000, 800, 400);
    init_pair(1, COLOR_YELLOW, COLOR_BLACK);
    init_pair(2, COLOR_GREEN, COLOR_BLACK);
    init_pair(3, COLOR_BLACK, COLOR_WHITE);


    keypad(stdscr, TRUE);
    cbreak();
    noecho();

    make_ui();

    loadPloggSettings();        // load up our basic settings

    if (argc <= 1) {
        // We weren't passed any song locations, so we'll use the playlist
        load_music_list();
    } else {
        // exit(1);
        /* We have been passed a list of songs to play (this time only) */
        oneShotPlaylist = 1;
        nrsongs = argc - 1;     // Number of songs passed
        tunes = (struct tune *)malloc(sizeof(struct tune)*nrsongs);
        bzero((void *)tunes, sizeof(struct tune)*nrsongs);

        for (i = 0; i < nrsongs; i++) {
            // sscanf(buf,"%[^#]#%d", b2, &(tunes[cnt].time));
            // tunes[cnt++].path = (char *)strdup(b2);
            fprintf(stderr, argv[i + 1]);
            fprintf(stderr, "**");
            tunes[i].time = 0;
            tunes[i].path = (char *)argv[i + 1];

            // Tag the songs as we go
            tunes[i].status|=TAGGED;
        }

        nrtagged=nrsongs;       // total number of tagged songs
        // playnextselected();     // start playing!
        playSelectedSong();

        // exit(0);
        // dynamic_music_list();
    }

    /* we must set the signals AFTER the configuration is read, or else
     * we'll cause a segv when the wc -l returns
     */

    // signal(SIGINT, finish);
    signal((int)finish, SIG_ACK);
    // signal(SIGCHLD, dostuff);
    signal((int)dostuff, SIG_ACK);

#ifdef LINUX_SOUND
    if ((mixer_fd = open("/dev/mixer", O_RDWR)) < 0) {
        fprintf(stderr, "Error opening /dev/mixer.\n");
        exit(1);
    }

    ioctl(mixer_fd, MIXER_READ(SOUND_MIXER_VOLUME), &volume);
    volume = volume & 0xff;
#endif

    refresh();
    refresh_screen();

    for(;;) {
        // if (!initialized) {
        /** I don't know why, but if we do a refresh before this loop
         * everything vanishes!  Fix Me XXX!!
         */
        // refresh_screen();
        //    initialized = TRUE;
        // }
        c = getch();
        action(c);
    }
}

/** Choose the proper player for the selected music type */
void choosePlayer(int choice) {
    int rc, status;         // return status... for use by a few things
    int l = strlen(tunes[choice].path);
    char testChoice[l+ 1];
    strcpy(testChoice, tunes[choice].path);

    // We have the check and see if there is a player running already
    if (playerpid != 0) {
        // Check the status of our player
        rc = waitpid(playerpid, &status, WNOHANG);

        if (rc >= 0) {
            // kill the existing player
            kill(playerpid, 9); /* yeah! BFG! */
            // signal(playerpid, SIGKILL);
        }
    }

    // Parse apart any arguments
    char workString[128];

    if (strstr(strupr(testChoice), ".OGG") > 0) {
        strcpy(workString, oggArgs);
        // strcat(workString, oggArgs);
    } else {
        strcpy(workString, mpgArgs);
        // strcat(workString, mpgArgs);
    }

    // fprintf(stderr, "Our workString is: %s \n", workString);

    // Count number of tokens
    int count = 0;
    char temp_list[25][256];
    char *countTok, *tok;
    char splat[25] = {0};
    int totalTokens = 0;
    countTok = strtok (workString," -");
    // fprintf(stderr, "Begin parsing arguments\n");
    while (countTok != NULL) {
        totalTokens++;
        //fprintf(stderr, "Starting strncpy.");
        strcpy(splat, "-");
        // fprintf(stderr, "Initial splat: %s \n", splat);
        strcat(splat, countTok);
        // fprintf(stderr, "Second splat: %s \n", splat);
        // strcpy(arg_list[totalTokens], splat);
        strcpy(temp_list[totalTokens], splat);
        // fprintf(stderr, "Post process: %s \n", temp_list[totalTokens]);
        countTok = strtok(NULL, " -");
    }

    //fprintf(stderr, "Total tokens: %i \n", totalTokens);

    char *arg_list[totalTokens + 1];

    if (strstr(strupr(testChoice), ".OGG") > 0) {
        strcpy(workString, oggArgs);
        arg_list[0] = oggPath;
    } else {
        strcpy(workString, mpgArgs);
        arg_list[0] = mpgPath;
    }

    int j;

    for (j = 1; j <= totalTokens; j++) {
        // strcpy(arg_list[j], temp_list[j]);
        // fprintf(stderr, "Token %i is %s \n", j, temp_list[j]);
        arg_list[j] = temp_list[j];
        // fprintf(stderr, "Token %i is ->%s<- \n", j, arg_list[j]);
    }

    arg_list[totalTokens + 1] = tunes[choice].path;
    arg_list[totalTokens + 2] = NULL;


    playerpid = spawnv(P_NOWAIT, arg_list[0], arg_list);

}

/** Determine how many lines are in a given file, and then put the titles in fileName */
void countLines(char *fileName) {
    FILE *fid;
    // int newlines;
    int c;
    int done = 0;
    int lastchar = 0;

    nrsongs = 0;

    // Open the file in read-only mode
    fid = fopen(fileName, "r");

    if (!fid) {
        fprintf(stderr, "I'm sorry, I can't find the playlist (called 'list').\n");
        fprintf(stderr, "Please check to make certain you have created one.");
        exit(0);
    }

    while(done == 0) {
        c = fgetc(fid);

        if(c == '\n') {
            nrsongs++;
        }

        if(c == EOF) {
            if(lastchar != '\n') {
                nrsongs++;
            }
            done = 1;
        }
        lastchar = c;
    }
    // Make certain to close the file
    fclose(fid);

    // This should return the total number of lines in our file
    return;
}

static void load_music_list() {
    FILE *in;
    char buf[1024],b2[1024];
    int cnt = 0;

    /* this is a simple format. just load a list of paths and be happy with
     * it (for now, improvements comes probably later. one idea is to
     * support some kind of detection for which cd things are on, and
     * make the program check a few possible paths etc)
     */

    /* count number of tunes first!
     */

    countLines(musiclist);

    // assert(in = fopen(musiclist, "r"));
    in = fopen(musiclist, "r");

    if (!in) {
        fprintf(stderr, "No music list found.");
        exit(0);
    }

    tunes = (struct tune *)malloc(sizeof(struct tune)*nrsongs);
    bzero((void *)tunes, sizeof(struct tune)*nrsongs);

    while(!feof(in)) {
        if(!fgets(buf, 1024, in)) {
            break;
        }
        buf[strlen(buf)-1] = 0; /* lose the trailing \n */
        sscanf(buf,"%[^#]#%d", b2, &(tunes[cnt].time));
        tunes[cnt++].path = (char *)strdup(b2);
    }

    fclose(in);
}

/** Read through the plogg.ini file and try to grab the locations of
 * our needed and optional programs:
 * ogg123 (required for ogg playback)
 * mpg123 (required for mp3 playback)
 * vorbiscomment (optional for information-tag support)
 */
void loadPloggSettings() {
    FILE *in;
    char buf[1024],b2[1024], name[1024];
    char *token1, token2;

    // int cnt = 0;
    int i;
    int eqSign = 0;

    // assert(in = fopen("plogg.ini", "r"));
    in = fopen("plogg.ini", "r");

    if (!in) {
        fprintf(stderr, "The plogg.ini file was not found.");
        return;      // This is a non-fatal error as we'll just use defaults
    }

    while (!feof(in)) {
        if (!fgets(buf, 1024, in)) {
            break;
        }

        // fprintf(stderr, buf);

        eqSign = -1;

        for (i = 0; i < strlen(buf); i++) {
            if (buf[i] == '=') {
                eqSign = i;
                // fprintf(stderr, "Bingo!  We have an equals sign at position %i \n", i);
            }

            if (i > eqSign && eqSign > -1) {
                // fprintf(stderr, "*%c* ", buf[i]);
                name[i - (eqSign + 1)] = buf[i];
                name[i - eqSign] = NULL;
            }
        }

        int len = strlen(name) - 1;     // length of our name variable

        if (strstr(buf, "OGGPLAYER=") > 0) {
            strncpy(oggPath, name, len);
            // strcpy(oggPath, name);
            // fprintf(stderr, "Our name is ->%s<- \n", name);
            // fprintf(stderr, "oggPath of *%s* \n", oggPath);
        }

        if (strstr(buf, "OGGARGS=") > 0) {
            strncpy(oggArgs, name, len);
        }

        if (strstr(buf, "MP3PLAYER=") > 0) {
            strncpy(mpgPath, name, len);
        }

        if (strstr(buf, "MP3ARGS=") > 0) {
            strncpy(mpgArgs, name, len);
        }

        if (strstr(buf, "OGGCOMMENT=") > 0) {
            // fprintf(stderr, "We're left with %s \n", name);
        }

        if (strstr(buf, "MP3COMMENT=") > 0) {
            // fprintf(stderr, "We're left with %s \n", name);
        }
    }

    fclose(in);
}

/** Launch a player to deal with the currently selected song */
void playSelectedSong() {
    int i;

    /* If there is already a player going, we'll stop it first */
    if(playerpid) {
        signal(SIGCHLD, SIG_IGN);
        signal(playerpid, SIG_IGN);
        stop_playing();
    }

    choosePlayer(tunenr);

    startedtune = tunenr;

    starttime = time(0);
    wclear(top);
    wattron(top, COLOR_PAIR(2));
    wattron(top, A_BOLD);
    waddstr(top, modify_name(tunes[tunenr].path));
    wattroff(top, A_BOLD);
    mvwaddstr(top, 0, COLS-6, " 0:00");
    wattroff(top, COLOR_PAIR(2));
    wmove(top, 1, 0);
    for(i=0; i<COLS; i++) {
        waddch(top, '-');
    }

    wrefresh(top);
    signal(SIGALRM, alarmhandler);
    // signal(alarmhandler, SIG_ACK);
    alarm(1);

}

void drawTitle() {
}
